"""Tests for imaging"""
import unittest
import logging
import protonn
import numpy as np

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class Tests(unittest.TestCase):

    def test_channels(self):
        logger.info("test image channels")
        img = np.random.random((80, 100, 3)).astype(np.float32)
        shape1 = img.shape
        logger.info("original shape: {}".format(img.shape))
        img = protonn.data.imaging.misc.make_channels_first(img)
        logger.info("new shape: {}".format(img.shape))
        img = protonn.data.imaging.misc.make_channels_last(img)
        logger.info("resored shape: {}".format(img.shape))
        shape2 = img.shape
        assert shape1 == shape2

    def test_crop_pad(self):
        a = np.ones((6, 6, 6))
        shape = (3, 4, 8)
        result = protonn.data.imaging.misc.set_size(a, shape)
        self.assertEqual(shape, result.shape)



#  this is assuming chainer format
import pandas
import os
import json


def load_single_run(path):
    df = pandas.read_json(os.path.join(path, "log"))
    with open(os.path.join(path, "params.json")) as f:
        d = json.load(f)
    time_elapsed = df["elapsed_time"]
    time_epoch = [y - x for x, y in zip(time_elapsed, time_elapsed[1:])]
    time_epoch = [df["elapsed_time"][0]] + time_epoch
    df["epoch_time"] = time_epoch
    for key in d:
        try:
            df[key] = d[key]
        except:
            print("could not export key ", key)
        # TODO: flatten hierarchical metadata entries
    return df


def load_multiple_runs(path):
    experiments = []
    for d in sorted(os.listdir(path)):
        try:
            experiments.append(load_single_run(os.path.join(path, d)))
        except Exception as e:
            print("could not load ", d)
            pass
            # raise e
    return experiments



"""Tests for utls"""
import unittest
import logging
# import protonn
from protonn.utils import save_data_json, get_time_str, save_code

logger = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    def test_timestamp(self):
        stamp = get_time_str()
        logger.info("test_timestamp: " + stamp)

    def test_save_code(self):
        save_code("/tmp/protonn/saved_code")

    def test_main(self):
        logger.info("test save json")
        data = {"i": 34}
        save_data_json(data, "/tmp/protonn/data.json")



"""Tests for parameters"""

import unittest
import logging

import protonn
from protonn.parameters.core import view, observe, dump

logger = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    def test_observe(self):
        logger.info("test view")

        @view
        def my_scope():
            x = 5
            return x
        my_scope()

        logger.info("test observe")
        y = 23
        # name = observe(y)
        # assert name == "y"
        dump("/tmp/protonn")

    def test_observe2(self):

        @view
        def my_scope():
            class Spam:
                my_very_important_param = None
            Spam.my_very_important_param = 5  # type: Observed
            Spam.my_very_important_param = 10
            return Spam.my_very_important_param

        self.assertIn('Spam.my_very_important_param', protonn.parameters._parameters)
        self.assertEqual(protonn.parameters._parameters['Spam.my_very_important_param'], None)
        my_scope()
        self.assertEqual(protonn.parameters._parameters['Spam.my_very_important_param'], 10)



import code
import signal
import traceback


def debug(sig, frame):
    """Interrupt running process, and provide a python prompt for
    interactive debugging."""
    d = {'_frame': frame}         # Allow access to frame object.
    d.update(frame.f_globals)  # Unless shadowed by global
    d.update(frame.f_locals)

    i = code.InteractiveConsole(d)
    message = "Signal received : entering python shell.\nTraceback:\n"
    message += ''.join(traceback.format_stack(frame))
    i.interact(message)


def print_stack_on_signal():
    # signal.signal(signal.SIGUSR1, debug)  # Register handler
    signal.signal(signal.SIGINT, debug)  # Register handler



"""Configuration support for protonn

Config files are expected to be found in the .vecto folder in user's home.
The format is the same as jupyter notebooks
"""

from traitlets.config.loader import load_pyconfig_files
import os.path


def load_config():
    default_dir = os.path.expanduser("~/.local/protonn/")
    if os.path.isfile(os.path.join(default_dir, 'config.py')):
        c = load_pyconfig_files(['config.py'], default_dir)
        return c
    else:
        # TODO: create default config
        raise RuntimeError('configuration file not found, please create one in ~/.local/protonn/config.py')



import logging
import os
import platform
import sys
from pathlib import Path

import yaml

from protonn.utils import get_time_str


def parse_float(dic, key):
    if key in dic:
        if isinstance(dic[key], str):
            dic[key] = float(dic[key])


def load_yaml_file(path):
    with open(path, "r") as cfg:
        # TODO: isn't this the place to support fancy floats?
        data = yaml.load(cfg, Loader=yaml.SafeLoader)
    return data


# TODO: need better name
def load_yaml_config(path_config):
    params_user = load_yaml_file(path_config)
    parse_float(params_user, "initial_lr")
    parse_float(params_user, "max_lr")
    parse_float(params_user, "eps")
    parse_float(params_user, "beta1")
    parse_float(params_user, "beta2")
    return params_user


class BaseConfig(dict):
    def __init__(self, name_task, cluster_env, param_path=None):
        self["name_task"] = name_task
        if int(os.environ["NUM_GPUS_PER_NODE"]) > 0:
            self["devices"] = -1
            self["accelerator"] = "gpu"
        else:
            self["devices"] = 1
            self["accelerator"] = "cpu"
        if len(sys.argv) < 2 and param_path is None:
            print("run main.py config.yaml")
            print("or")
            print("run main.py logs/path/to/snapshot/epoc10_step42000")
            exit(-1)
        if param_path is None:
            path = Path(sys.argv[1])
        else:
            path = Path(param_path)
        self.set_defaults()
        self._is_master = cluster_env.global_rank() == 0
        self.read_from_yaml_and_set_default(path, name_task)
        self.add_distributed_info(cluster_env.world_size())
        self.maybe_create_unique_path()
        # TODO(vatai): create base trainer
        print("WARRNING: seed not set.  This will be implemented in trainer.base class")
        cluster_env.barrier()

    def get_run_folder(self):
        timestamp = self["timestamp"][:-3]
        hostname = platform.node().split(".")[0]
        workers = self["cnt_workers"]
        # TODO: this might be dynamic
        lr = self["max_lr"] * self["cnt_workers"]
        seed = self["seed"]
        run_folder = f"{timestamp}_w{workers}_lr{lr:.4f}_s{seed}_{hostname}"
        # TODO: make this trully unique
        return run_folder

    def maybe_create_unique_path(self):
        if self["create_unique_path"]:
            self["path_results"] = os.path.join(self["path_results"], self["name_project"])
            # TODO: it's hacky, but for the time being for langmo
            # ideally should be a list of names to concat into path
            if "model_name" in self:
                self["path_results"] = os.path.join(self["path_results"], self["model_name"])
            # TODO: extract nicemodel name from metadata
            # model_name = self["model_name"].split("/")[-1]
            # self["path_results"] = os.path.join(self["path_results"], model_name)
            run_dir = self.get_run_folder()
            self["path_results"] = os.path.join(self["path_results"], run_dir)
        else:
            # TODO: chech if the folder is empty
            pass
        if "WANDB_MODE" not in os.environ or os.environ["WANDB_MODE"].lower() != "disabled":
            if self._is_master:
                path_wandb = Path(self["path_results"]) / "wandb"
                path_wandb.mkdir(parents=True, exist_ok=True)

    # TODO: we have near identical method in langmo
    def read_from_yaml_and_set_default(self, path, name_project):
        self["name_project"] = name_project
        self["timestamp"] = get_time_str()
        _logger = logging.getLogger(__name__)
        user_config = load_yaml_config(path)
        for key, value in user_config.items():
            if key not in self.defaults and key not in self.required_options and key != "suffix":
                raise RuntimeError(f"got unexpected key in user config\t{key}: {value}")
            # print(key, value)
        for key in self.required_options:
            if key not in user_config:
                raise RuntimeError(f"required key not in config {key}")
        for key, value in self.defaults.items():
            if key not in user_config:
                if self._is_master:
                    _logger.warning(f"setting parameter {key} to default value {value}")
                self[key] = value

        self.update(user_config)

    def add_distributed_info(self, cnt_workers):
        self["cnt_workers"] = cnt_workers
        # TODO: we remove this because accumulate batches is now dynamic
        # batch_size = self["batch_size"]
        # acc_batches = self["accumulate_batches"]
        # self["batch_size_effective"] = batch_size * cnt_workers * acc_batches

    def set_defaults(self):
        self.defaults = dict()
        self.defaults["seed"] = 0
        self.defaults["create_unique_path"] = True
        self.required_options = set()
